# Copyright 2013, 2014, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.

use strict;
use Carp;
use base 'Exporter';
use Math::Matrix;

use Math::PlanePath::GcdRationals;
*gcd = \&Math::PlanePath::GcdRationals::_gcd;

# uncomment this to run the ### lines
# use Smart::Comments;

our @EXPORT_OK = ('U_3x3','A_3x3','D_3x3',
                  'M1_3x3','M2_3x3','M3_3x3','T_3x3',
                  'U_2x2','A_2x2','D_2x2',
                  'M1_2x2','M2_2x2','M3_2x2','T_2x2',
                  'matrix_name',
                  'mm_2x2_to_3x3',
                  'mvec_to_pq','mvec_to_str',
                  'ternary_1s_reverse',
                  'ternary_1s_unreverse',
                  'is_integer',
                  'pq_acceptable',
                  'AB_acceptable',
                  'line_is_horizontal','line_is_diagonal','line_sort_ends',
                  'line_end_clip',,
                  'printwrap','printwrap_newline','printwrap_indent',
                 );
our @EXPORT_TAGS = (':all' => \@EXPORT_OK);


=pod

# Pari:
 u = [ 1,-2, 2;  2, -1, 2; 2, -2, 3];
 a = [ 1, 2, 2;  2,  1, 2; 2,  2, 3];
 d = [-1, 2, 2; -2,  1, 2; -2, 2, 3];

=cut

use constant U_3x3 => Math::Matrix->new ([1, -2, 2],    # Hall "L"
                                         [2, -1, 2],
                                         [2, -2, 3]);
use constant A_3x3 => Math::Matrix->new ([1, 2, 2],     # Hall "U"
                                         [2, 1, 2],
                                         [2, 2, 3]);
use constant D_3x3 => Math::Matrix->new ([-1, 2, 2],    # Hall "R"
                                         [-2, 1, 2],
                                         [-2, 2, 3]);

use constant M1_3x3 => Math::Matrix->new ([2, 1, -1], # misprint 2,2,1 in v1
                                          [-2, 2, 2],
                                          [-2, 1, 3]);
use constant M2_3x3 => Math::Matrix->new ([2, 1, 1],  # misprint 2,2,1 in v1
                                          [2, -2, 2],
                                          [2, -1, 3]);
use constant M3_3x3 => Math::Matrix->new ([2, -1, 1], # misprint 2,-2,1 in v1
                                          [2, 2, 2],
                                          [2, 1, 3]);

use constant T_3x3 => Math::Matrix->new ([-2, 3, 3],
                                         [-6, 2, 6],
                                         [-6, 3, 7]);

use constant U_2x2 => Math::Matrix->new ([2, -1],   # Hall "L"  det=+1
                                         [1, 0]);
use constant A_2x2 => Math::Matrix->new ([2, 1],    # Hall "U"  det=-1
                                         [1, 0]);
use constant D_2x2 => Math::Matrix->new ([1, 2],    # Hall "R"  det=+1
                                         [0, 1]);

use constant M1_2x2 => Math::Matrix->new ([1, 1],
                                          [0, 2]);
use constant M2_2x2 => Math::Matrix->new ([2, 0],
                                          [1, -1]);
use constant M3_2x2 => Math::Matrix->new ([2, 0],
                                          [1, 1]);

use constant T_2x2 => Math::Matrix->new ([1, 3],
                                         [0, 2]);

use constant X1_2x2 => U_2x2() * U_2x2();
use constant X2_2x2 => M1_2x2() * M2_2x2();


# $m is a 2x2  Math::Matrix, return its 3x3 form Math::Matrix
sub mm_2x2_to_3x3 {
  my ($m) = @_;
  my ($width, $height) = $m->size;
  $width == 2 && $height == 2
    or croak "mm_2x2_to_3x3() not a 2x2 matrix";

  # subst to abcd otherwise per Palmer, Ahuja, Tikoo
  #
  my $a = $m->[0][0]; my $b = $m->[0][1];
  my $c = $m->[1][0]; my $d = $m->[1][1];

  return Math::Matrix->new ([ (($a*$a - $c*$c) - ($b*$b - $d*$d)) / 2,
                              $a*$b - $c*$d,
                              (($a*$a - $c*$c) + ($b*$b - $d*$d)) / 2 ],

                            [ $a*$c - $b*$d,
                              $a*$d + $b*$c,
                              $a*$c + $b*$d ],

                            [ (($a*$a + $c*$c) - ($b*$b + $d*$d)) / 2,
                              $a*$b + $c*$d,
                              (($a*$a + $c*$c) + ($b*$b + $d*$d)) / 2 ]);

  #                            # d=0
  # aa - cc - bb + dd          aa - cc - bb
  # ab - cd                    ab
  # aa - cc + bb - dd          aa - cc + bb

  # U->A sign change b
  #


  # # as per Palmer, Ahuja, Tikoo
  # my $r = $m->[0][0]; my $s = $m->[0][1];
  # my $t = $m->[1][0]; my $u = $m->[1][1];
  #
  # return Math::Matrix->new ([ (($r*$r - $t*$t) - ($s*$s - $u*$u)) / 2,
  #                             $r*$s - $t*$u,
  #                             (($r*$r - $t*$t) + ($s*$s - $u*$u)) / 2 ],
  #
  #                           [ $r*$t - $s*$u,
  #                             $r*$u + $s*$t,
  #                             $r*$t + $s*$u ],
  #
  #                           [ (($r*$r + $t*$t) - ($s*$s + $u*$u)) / 2,
  #                             $r*$s + $t*$u,
  #                             (($r*$r + $t*$t) + ($s*$s + $u*$u)) / 2 ]);
}



sub is_integer {
  my ($n) = @_;
  return $n == int($n);
}
sub is_square {
  my ($n) = @_;
  my $s = int(sqrt($n));
  my $ss = $s*$s;
  if ($ss > $n) {
    $s--;
    $ss = $s*$s;
  }
  return $ss == $n;
}

BEGIN {
  my %matrix_name;

  my $mm_add = sub {
    my ($m, $name) = @_;
    $matrix_name{math_matrix_string($m)} ||= $name;
  };
  $mm_add->(U_2x2,'U');
  $mm_add->(A_2x2,'A');
  $mm_add->(D_2x2,'D');
  $mm_add->(M1_2x2,'M1');
  $mm_add->(M2_2x2,'M2');
  $mm_add->(M3_2x2,'M3');
  $mm_add->(T_2x2,'T');

  # products
  foreach my $pass (1, 2) {
    ### $pass
    ### %matrix_name
    my @keys = keys %matrix_name;
    foreach my $x (@keys) {
      my ($a,$b,$c,$d) = split /,/, $x;
      my $x_2x2 = Math::Matrix->new([$a,$b],[$c,$d]);
      my $x_name = $matrix_name{$x};
      foreach my $y (@keys) {
        my ($a,$b,$c,$d) = split /,/, $y;
        my $y_2x2 = Math::Matrix->new([$a,$b],[$c,$d]);
        my $y_name = $matrix_name{$y};
        $mm_add->($x_2x2*$y_2x2, "$x_name.$y_name");
      }
    }
  }

  # inverses
  {
    my @keys = keys %matrix_name;
    foreach my $x (@keys) {
      my ($a,$b,$c,$d) = split /,/, $x;
      my $x_2x2 = Math::Matrix->new([$a,$b],[$c,$d]);
      my $x_name = $matrix_name{$x};
      $mm_add->($x_2x2->invert, "($x_name)inv");
    }
  }

  # transposes
  {
    my @keys = keys %matrix_name;
    foreach my $x (@keys) {
      my ($a,$b,$c,$d) = split /,/, $x;
      my $x_2x2 = Math::Matrix->new([$a,$b],[$c,$d]);
      my $x_name = $matrix_name{$x};
      $mm_add->($x_2x2->transpose, "($x_name)transpose");
    }
  }

  # 3x3s
  {
    my @keys = keys %matrix_name;
    foreach my $x (@keys) {
      my ($a,$b,$c,$d) = split /,/, $x;
      my $x_2x2 = Math::Matrix->new([$a,$b],[$c,$d]);
      my $x_name = $matrix_name{$x};
      my $x_3x3 = mm_2x2_to_3x3($x_2x2);
      $mm_add->($x_3x3, $x_name);
    }
  }

  sub math_matrix_string {
    my ($m) = @_;

    my ($width, $height) = $m->size;
    $width == 2 && $height == 2
      or $width == 3 && $height == 3
        or croak "math_matrix_string() unrecognised matrix size";

    my @terms = map{@$_}@$m;
    return join(',',@terms);
  }

  sub matrix_name {
    my $key;
    if (@_ == 1) {
      my ($m) = @_;
      $key = math_matrix_string($m);
    } elsif (@_ == 4 || @_ == 9) {
      $key = join(',',@_);
    } else {
      croak "matrix_name() bad args";
    }
    return $matrix_name{$key} || '?';
  }

}

sub ternary_1s_reverse {
  my ($n) = @_;
  ### ternary_1s_reverse(): $n
  require Math::PlanePath::Base::Digits;
  my @digits = Math::PlanePath::Base::Digits::digit_split_lowtohigh($n,3);
  {
    require Math::PlanePath::GrayCode;
    Math::PlanePath::GrayCode::_digits_to_gray_reflected(\@digits,3);
  }
  # {
  #   my $rev = 0;
  #   foreach my $digit (reverse @digits) {  # high to low
  #     if ($rev) { $digit = 2 - $digit; }   # mutate array
  #     $rev ^= ($digit == 1);
  #   }
  # }
  return Math::PlanePath::Base::Digits::digit_join_lowtohigh(\@digits,3);
}
sub ternary_1s_unreverse {
  my ($n) = @_;
  ### ternary_1s_reverse(): $n
  require Math::PlanePath::Base::Digits;
  require Math::PlanePath::GrayCode;
  my @digits = Math::PlanePath::Base::Digits::digit_split_lowtohigh($n,3);
  Math::PlanePath::GrayCode::_digits_from_gray_reflected(\@digits,3);
  return Math::PlanePath::Base::Digits::digit_join_lowtohigh(\@digits,3);
}

sub pq_acceptable {
  my ($p,$q) = @_;
  unless (is_integer($p)) {
    ### not integer p ...
    return 0;
  }
  unless (is_integer($q)) {
    ### not integer q ...
    return 0;
  }
  return ($p > $q
          && $q >= 1
          && ($p % 2) != ($q % 2)
          && gcd($p,$q) == 1);
}
sub AB_acceptable {
  my ($A,$B) = @_;
  unless (is_integer($A)) {
    return 0;
  }
  unless (is_integer($B)) {
    return 0;
  }
  return ($A > 0
          && $B > 0
          && is_square($A*$A + $B*$B)
          && ($A % 2) == 1
          && ($B % 2) == 0
          && gcd($A,$B) == 1);
}

#------------------------------------------------------------------------------
# Math::Matrix things

# $m is a Math::Matrix which should be a 2x1 column vector [[$p],[$q]].
# Return the two values ($p,$q).
sub mvec_to_pq {
  my ($m) = @_;
  my ($rows,$cols) = $m->size;
  die unless $rows == 2 && $cols == 1;
  return ($m->[0][0], $m->[1][0]);
}

# $m is a Math::Matrix which should be a 2x1 column vector [[$p],[$q]].
# Return a string "$p,$q".
sub mvec_to_str {
  my ($m) = @_;
  my ($p,$q) = mvec_to_pq($m);
  return "$p,$q";
}

#------------------------------------------------------------------------------
# lines

# return true if the line x1,y1 to x2,y2 is horizontal, ie. $y1==$y2
sub line_is_horizontal {
  my ($x1,$y1, $x2,$y2) = @_;
  return $y1 == $y2;
}

# return true if the line x1,y1 to x2,y2 is diagonal, ie. slope +1
sub line_is_diagonal {
  my ($x1,$y1, $x2,$y2) = @_;
  return $x1-$y1 == $x2-$y2;
}

# return (x1,y1, x2,y2) with the ends swapped if necessary to make x1<x2, or
# if vertical x1=x2 then y1<y2
sub line_sort_ends {
  my ($x1,$y1, $x2,$y2) = @_;
  if ($x1 > $x2 || ($x1 == $x2 && $y1 > $y2)) {
    return ($x2,$y2, $x1,$y1);
  } else {
    return ($x1,$y1, $x2,$y2);
  }
}

# $x1,$y1 to $x2,$y2 is a line.
# Clip the $x2,$y2 end so that it is within $x_max,$y_max.
# Return the new $x2,$y2 values.
sub line_end_clip {
  my ($x1,$y1, $x2,$y2, $x_max,$y_max) = @_;
  ### _line_end_clip(): "$x1,$y1, $x2,$y2"

  my ($x2new, $y2new);
  if ($x2 > $x_max) {
    # x2 is outside LO to HI, clip to x2=HI and y2 set to corresponding
    my $xlen = $x2 - $x1;

    $x2new = $x_max;
    $y2new = ($y2*($x2new-$x1) + $y1*($x2-$x2new)) / $xlen;
    ### x clip: "to $x2new,$y2new"
  } else {
    $x2new = $x2;
    $y2new = $y2;
  }

  if ($y2new > $y_max) {
    my $ylen = $y2 - $y1;
    $y2new = $y_max;
    $x2new = ($x2*($y2new-$y1) + $x1*($y2-$y2new)) / $ylen;
    ### y clip: "to $x2new,$y2new   left ".($y2new-$y1)." right ".($y2-$y2new)
  }

  return ($x2new,$y2new);
}

#------------------------------------------------------------------------------
1;
__END__
