#!/usr/bin/perl -w

# Copyright 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.010;
use strict;
use List::Util 'max','min';
use Math::PlanePath::PythagoreanTree;
use Math::PlanePath::Base::Digits;

use FindBin;
use lib $FindBin::Bin;
use MyTriples 'U_2x2','A_2x2','D_2x2';
$| = 1;

print "  % generated by tools/$FindBin::Script\n";
print "  \\begin{scope}[font=\\footnotesize,inner sep=.25em]\n";

my $path = Math::PlanePath::PythagoreanTree->new(tree_type => 'UArD',
                                                 coordinates => 'PQ');
my @UAD_matrices = (U_2x2(),A_2x2(), D_2x2());

sub n_to_depth_and_offset {
  my ($n) = @_;
  my $depth = $path->tree_n_to_depth($n);
  my $n_lo = $path->tree_depth_to_n($depth);
  return ($depth, $n - $n_lo);
}
sub node_name {
  my ($depth,$offset) = @_;
  return "p$depth-$offset";
}
sub n_to_node_name {
  my ($n) = @_;
  my ($depth,$offset) = n_to_depth_and_offset($n);
  return "p$depth-$offset";
}

for (my $n = $path->n_start; ; $n++) {
  my ($depth,$offset) = n_to_depth_and_offset($n);
  last if $depth > 3;
  my ($p,$q) = $path->n_to_xy($n);
  my $name = n_to_node_name($n);


  my $is_2big = ($depth == 2 && 3 <= $offset && $offset < 6);
  my $is_2small = ($depth == 2 && !$is_2big);
  my $is_3small = ($depth == 3 && ($offset < 9 || $offset >= 18));
  my $is_3medium = ($depth == 3 && ((9 <= $offset && $offset < 12)
                                    || (15 <= $offset && $offset < 18)));
  my $is_3big = ($depth == 3 && 9 <= $offset && $offset < 18);
  my $is_small = $is_2small || $is_3small;

  my @digits = reverse   # high to low
    Math::PlanePath::Base::Digits::digit_split_lowtohigh($offset,3);
  while (@digits < $depth) { unshift @digits, 0; }
  @digits == $depth or die;
  my $pos = '';
  if ($depth == 0) {
    $pos = '0,0';
  } else {
    my $join = '';
    foreach my $i (1 .. $depth) {
      my $factor = sprintf '%+d', $digits[$i-1] - 1;
      my $spread = ($i>=2 && $is_small ? "spread${i}small"
                    : $i==3 && $is_3medium ? "spread${i}medium"
                    : "spread$i");
      $pos = "$factor*($spread)".$pos;
    }
    $pos = "\$$pos+(row$depth)\$";
  }

  if ($depth == 3 && $offset == 0) {
    print "  \\begin{scope}[font=\\scriptsize]\n";
  }

  if ($is_3small) {
    my $dots = ($digits[-1] == 1 ? '\\cdots' : '\\phantom{\\cdots}');
    print "  \\node ($name) at ($pos) {\$$dots\\vphantom{1}\$};\n";
  } else {
    print "  \\node ($name) at ($pos) {\$${p}{,}$q\$};\n";
  }

  if (defined (my $parent_n = $path->tree_n_parent($n))) {
    my ($parent_p,$parent_q) = $path->n_to_xy($parent_n);
    my $parent_name = n_to_node_name($parent_n);

    my @Gray_digits = @digits;
    {
      my $rev = 0;
      foreach my $i (0 .. $#digits) {     # high to low
        if ($rev) { $Gray_digits[$i] = 2 - $digits[$i]; }
        if ($digits[$i] == 1) { $rev = !$rev; }
      }
    }

    # high to low, of the Gray code
    my @matrices = map {$UAD_matrices[$_]} @Gray_digits;

    my $mvec = Math::Matrix->new([2],[1]);
    foreach my $m (@matrices) {
      $mvec = $m * $mvec;  # high to low on the left
    }
    $mvec->equal(Math::Matrix->new([$p],[$q]))
      or die "oops, depth=$depth offset=$offset",
      " digits=",join(',',@digits),
      " Gray=",join(',',@Gray_digits),
      " by matrix $mvec want $p,$q";

    my $m = $matrices[-1];   # last applied
    my $parent_mvec = Math::Matrix->new([$parent_p],[$parent_q]);
    $mvec->equal($m * $parent_mvec) or die;

    my $m_name = matrix_name ($m);
    my $base_pos = ($depth == 3 ? '\\MyBasePosThree' : '\\MyBasePos');

    if ($depth==3
        && ($offset < 9 || $offset==10 || $offset==16 || $offset >= 18)) {

      # m name in middle of line (in depth=3)
      my @style;
      if ($digits[-1] != 1) {
        my $xshift = '.3em';
        my $sign = $digits[-1] == 0 ? '-' : '';
        push @style, "xshift=$sign$xshift";
      }
      my $style = join(',',@style);
      print "  \\node at (\$($parent_name)!$base_pos!($name)\$) [$style] {$m_name};\n";
      print "  \\draw ($parent_name) -- (\$($parent_name)!\\MyTopPos!($name)\$) ",
        "(\$($parent_name)!\\MyBottomPos!($name)\$) -- ($name);\n";

    } else {
      # m name beside line

      my @style;
      if ($depth == 3) {
        push @style, ($digits[-1] == 0   ? 'left=.25em'
                      : $digits[-1] == 1 ? 'right'
                      :                    'right=.4em');
      } else {
        push @style, $digits[-1] == 0 ? 'base left' : 'base right';
      }
      push @style, 'inner xsep=.1em';

      # shallow lines space up a bit
      if ($depth == 1 && $digits[-1] != 1) { push @style, "yshift=.4ex"; }
      if ($depth == 2 && ($offset==3||$offset==5)) { push @style,"yshift=.3ex";}
      my $style = join(',',@style);
      print "  \\draw ($parent_name) -- ($name);\n";
      print "  \\node at (\$($parent_name)!$base_pos!($name)\$) [$style] {$m_name};\n";
    }
  }
}
print "  \\end{scope}\n";
print "  \\end{scope}\n";

print "  % end generated\n";
exit 0;
