#!/usr/bin/perl -w

# Copyright 2013, 2014, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


# Create web page picture image.png.

use 5.006;
use strict;
use warnings;
use File::chdir;
use File::stat;
use File::Slurp;
use File::Temp;
use IPC::Run;
use List::Util 'max','min';

use FindBin;
use lib $FindBin::Bin;


my $AlphaBits = 4;   # 1,2,4

my $version = 0;
foreach my $arg (@ARGV) {
  if ($arg =~ /^VERSION=(.*)/) {
    $version = $1;
  }
}

my $program_filename = File::Spec->catfile($FindBin::Bin,
                                           $FindBin::Script);
my $png_filename = File::Spec->catfile($FindBin::Bin,
                                       File::Spec->updir,
                                       'image.png');

my $temp = File::Temp->newdir;
my $tempdir = $temp->dirname;
local $CWD = $tempdir;


my $triples_tex_filename = File::Spec->catfile($FindBin::Bin,
                                               File::Spec->updir, 
                                               'triples.tex');
my $image_str = File::Slurp::read_file($triples_tex_filename);
$image_str =~ /^%% MY-BEGIN-IMAGE(.*)^%% MY-END-IMAGE/ms
  or die "oops, MY-BEGIN-IMAGE not matched";
$image_str = $1;

$image_str = <<"HERE";
% extracted from triples.tex -- DO NOT EDIT
\\documentclass[a4paper]{article}
\\pagestyle{empty}
\\usepackage{tikz}
\\tikzset{>=latex,       % arrowhead style
          font=\\small}
\\begin{document}
$image_str
\\end{tikzpicture}
\\end{document}
HERE

print $image_str;
File::Slurp::write_file('image.tex', $image_str);

IPC::Run::run(['latexmk','-pdf','image.tex'], '<',File::Spec->devnull);

IPC::Run::run(['pdfcrop',
               '--verbose',
               '--margins','10',
               'image.pdf',
               'image-cropped.pdf'],
              '<', File::Spec->devnull)
  or die;
rename 'image-cropped.pdf', 'image.pdf' or die;

{
  my @command = ('gs',
                 '-q',  # quiet, cf -dQUIET
                 '-dNOPAUSE',
                 '-r100',
                 '-sDEVICE=pnggray',
                 "-dTextAlphaBits=$AlphaBits",     # grey scale edge blurring
                 "-dGraphicsAlphaBits=$AlphaBits",
                 "-sOutputFile=$png_filename",
                 'image.pdf'); 
  print join(' ',@command),"\n";
  IPC::Run::run(\@command, '<', File::Spec->devnull)
      or die;
}
print "\n";

# PNG spec 11.3.4.2 suggests RFC822 (or rather RFC1123) for CreationTime
use constant STRFTIME_FORMAT_RFC822 => '%a, %d %b %Y %H:%M:%S %z';
my $stat = File::stat::stat($program_filename)
  or die "Oops, cannot stat $program_filename: $!";
my $creation_time = POSIX::strftime(STRFTIME_FORMAT_RFC822,
                                    localtime($stat->mtime));

my $copyright = File::Slurp::read_file($program_filename);
$copyright =~ /^# (Copyright.*)/m or die;
$copyright = $1;

pngtextadd($png_filename, 'Author',    'Kevin Ryde');
pngtextadd($png_filename, 'Generator', "tools/$FindBin::Script");
pngtextadd($png_filename, 'Title',     'Trees of Primitive Pythagorean Triples - sample image');
pngtextadd($png_filename, 'Creation Time', $creation_time);
pngtextadd($png_filename, 'Copyright', $copyright);
pngtextadd($png_filename, 'Homepage',  'http://user42.tuxfamily.org/triples/index.html');
# pngtextadd($png_filename, 'Version',   $version);

print "uncompressed ",-s $png_filename,"\n";
IPC::Run::run(['optipng','-quiet','-o5',$png_filename]);
print "compressed   ",-s $png_filename,"\n";


# IPC::Run::run(['xpdf','image.pdf']);
# IPC::Run::run(['xzgv',$png_filename]);

exit 0;

# add text to the png image in $png_filename
sub pngtextadd {
  my ($filename, $keyword, $value) = @_;
  system('pngtextadd', "--keyword=$keyword", "--text=$value", $filename) == 0
    or die "system(pngtextadd)";
}
