#!/usr/bin/perl -w

# Copyright 2019, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.


# Usage: perl test-gen-output.pl filename.tex ...
#
# Look in filename.tex for parts generated by tools/gen-*.pl programs,
#
#     % generated by tools/gen-foo.pl 123
#     ...
#     % end generated
#
# Run the program, with any options, and compare its output to what
# filename.tex contains.  A filename:linenum: error message is given for any
# mismatches, and a diff -u to show perhaps only something minor different.
#
# The aim is to detect regressions in the programs, or perhaps global
# changes made in filename.tex which the programs ought to have too.
# Programs can be any executables, though currently all are Perl .pl.
#
# Programs are currently run with current directory set to the directory of
# its originating filename.tex, so that the program name is tool/gen-foo.pl
# relative to that.  It's possible generator programs were run from their
# own directory ("tools") when first created.  They can take that into
# account by going "FindBin" for helper modules etc in their own directory.
#
# Bugs:
#
# There's a hard-coded list of programs and options which are too slow to
# want to run every time.  This is usually generating sample data values in
# some exhaustive way, ready to compare against formulas.  Should have some
# way to indicate slowness in filename.tex and skip.
#

use 5.010;
use strict;
use warnings;
use File::chdir;
use File::Basename;
use File::Slurp;
use File::Spec;
use IPC::Run;
use List::Util 'sum';
$|=1;

my $verbose = 1;
my $bad = 0;

my %slow_commands =
  (
   # these sizes too big to run every time

   # dragon.tex
   'tools/gen-GJA-values.pl 16' => 1,
   'tools/gen-GRQtotal-values.pl 14' => 1,
   'tools/gen-TAW-values.pl 10' => 1,
   'tools/gen-blob-area-crl.pl 16' => 1,
   'tools/gen-blob-sample-values.pl 16' => 1,
   'tools/gen-convex-hull-boundary-values.pl 30' => 1,
   'tools/gen-edge-count-samples.pl 16' => 1,
   'tools/gen-hanging-values.pl 16' => 1,
   'tools/gen-path-boundary-diameter.pl 10' => 1,
   'tools/gen-ffixed-ones-mean-plot.pl 44' => 1,  # about 20 mins for 2^44

   # alternate.tex
   'tools/gen-TW-values.pl 8' => 1,
   'tools/gen-TAEdgeCount-values.pl 14' => 1,
   'tools/gen-TAW-values.pl 8' => 1,

   # terdragon.tex
   'tools/gen-centroid-boundary-samples.pl 12' => 1,
   'tools/gen-centroid-area-values.pl 10' => 1,

   # r5dragon.tex
   'tools/slow-centroid-boundary-samples.pl 11' => 1,
   'tools/slow-GJ-centroid-values.pl 9' => 1,
   'tools/gen-hull-values.pl 60' => 1,
   'tools/gen-Qhull-values.pl 60' => 1,
   'tools/gen-QADegCount-values.pl 6' => 1,
   'tools/gen-QAdiameter-values.pl 4' => 1,
   'tools/gen-QAW-values.pl 4' => 1,
   'tools/gen-QTW-values.pl 4' => 1,
   'tools/gen-StarRepW-values.pl big' => 1,
   'tools/gen-JADegCount-values.pl 7' => 1,
   'tools/gen-JAdiameter-values.pl 5' => 1,

   # c-curve.tex
   'tools/gen-triangle-regions-samples.pl 24' => 1,
   'tools/gen-Enc-values.pl 22' => 1,
   'tools/gen-Q-values.pl 12' => 1,
   'tools/gen-QD-values.pl 20' => 1,
  );

sub virtual_time {
  # total of ($user,$system,$cuser,$csystem)
  # the gen programs run count into cuser and csystem
  return sum(times());
}

foreach my $tex_filename (@ARGV) {
  my $tex_str = File::Slurp::slurp($tex_filename);

  my $tex_dirname = File::Basename::dirname($tex_filename);
  {
    my @dirs = File::Spec->splitdir($tex_dirname);
    if ($dirs[-1] eq 'devel') {
      $tex_dirname = File::Spec->catdir($tex_dirname, File::Spec->updir);
    }
  }

  local $CWD = $tex_dirname;
  print "$tex_filename, run in directory $CWD\n";

  my %gen_seen;
  my $count = 0;

  while ($tex_str =~
    m{^ *% generated by (?<command>[^\n]*).*?(?<end>% end generated\n|\Z)}smg) {
    my $command    = $+{'command'};
    my $tex_output = $&;
    my $end        = $+{'end'};
    my $linenum = pos_linenum($tex_str, $-[0]);
    if ($end eq '') {
      print "$tex_filename:$linenum: missing \"end generated\"\n";
      $bad = 1;
      next;
    }
    $count++;
    if ($verbose) {
      print "$command, output ",count_lines($tex_output)," lines\n";
    }
    my $command_is_reduced = 0;
    if ($slow_commands{$command}) {
      if ($command =~ /(.*?) /) {
        $command = $1;
        $command_is_reduced = 1;
        print "  reduce due to slow: $command\n";
      } else {
        print "  skip due to slow\n";
        next;
      }
    }

    my @command = split /[ \t]+/, $command;
    $gen_seen{$command[0]} = 1;
    $command[0] = "./$command[0]";
    my $t1 = virtual_time();
    my ($stdout, $stderr);
    unless (IPC::Run::run(\@command, '>',\$stdout, '2>',\$stderr)
            && $stderr eq '') {
      print "Error running $command: $?";
      print $stdout;
      print $stderr;
      $bad = 1;
      next;
    }
    my $elapsed = virtual_time() - $t1;
    if ($elapsed > 10) {  # seconds
      printf "  elapsed virtual time %.1f\n", $elapsed;
    }
    if (!$command_is_reduced
        && $tex_output ne $stdout) {
      print "$tex_filename:$linenum: different output $command\n";
      print "  tex length     ",length($tex_output),"\n";
      print "  program length ",length($stdout),"\n";
      diff_strings($tex_output,$stdout);
      $bad = 1;
      next;
    }
  }
  print "$count generator outputs\n";

  # Run otherwise unused gen programs.
  #
  foreach my $filename (glob("gen-*.pl")) {
    next if $gen_seen{$filename};
    if ($verbose) {
      print "$filename\n";
    }
    my @command = ("./$filename");
    my ($stdout, $stderr);
    unless (IPC::Run::run(\@command, '>',\$stdout, '2>',\$stderr)
            && $stderr eq '') {
      print "Error running ",join(' ',@command),": $?";
      print $stdout;
      print $stderr;
      $bad = 1;
      next;
    }
  }
  # EXIT=0; for i in tools/gen*.pl; do echo "$$i"; perl "$$i" >/dev/null || EXIT=1; done; exit $$EXIT
}


#------------------------------------------------------------------------------

# Print a "diff -u" of two strings $old, $new.
sub diff_strings {
  my ($old, $new) = @_;
  require File::Temp;
  my $old_fh = File::Temp->new;
  my $new_fh = File::Temp->new;
  print $old_fh $old or die;
  print $new_fh $new or die;
  IPC::Run::run(['diff', '-u', $old_fh->filename, $new_fh->filename]);
}

# Return the line number of character position $pos in $str.
# If $pos is undef or not supplied, use pos($str).
sub pos_linenum {
  my ($str, $pos) = @_;
  $pos //= pos($_[0]);
  $str = substr($str, 0, $pos);
  return 1 + scalar($str =~ tr/\n//);
}

sub count_lines {
  my ($str) = @_;
  my $count = ($str =~ tr/\n/\n/);
  if ($str =~ /[^\n]$/) { $count++; }
  return $count;
}

#------------------------------------------------------------------------------
exit ($bad ? 1 : 0);
